/*
 * gw/mms-appl.c - wapbox application layer implements MMS behaviour
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <arpa/inet.h>
#include <string.h>

#include "gwlib/gwlib.h"
#include "xml_shared.h"
#include "wml_compiler.h"
#include "wap/wap.h"
#include "mms-appl.h"
#include "wap/wsp_strings.h"
#include "wap-error.h"
#include "urltrans.h"
#include "mms/mms_pdu.h"
#include "gw/smsc/smsc_at2.h"


/* to start/stop dialup bearer service without having to go
   through its web interface */
extern SMSCConn *modemConn;

typedef enum {
#define STATE_NAME(name) name,
#define ROW(state, event, condition, action, next_state)
#include "mms/mms_receiver_states.def"

#define STATE_NAME(name) name,
#define ROW(state, event, condition, action, next_state)
#include "mms/mms_sender_states.def"

    MMSState_count
} MMSState;




#define HTTP_MAX_RETRIES    30
#define HTTP_RETRY_DELAY    10 /* in sec. */

/*
 * Give the status the module:
 *
 *	limbo
 *		not running at all
 *	running
 *		operating normally
 *	terminating
 *		waiting for operations to terminate, returning to limbo
 */
static enum {
    limbo, running, terminating
} run_status = limbo;


/*
 * The queue of incoming events.
 */
static List *queue = NULL;
/* 
 * Incoming MMS messages 
 */
static List *incoming_mms = NULL;

/* machines for sending and recieving */
static List *clientMachines = NULL;
static List *clientSendMachines = NULL;
static List *http_requests = NULL; /* the outbound HTTP request queue */

/*
 * HTTP caller identifier for application layer.
 */
HTTPCaller *mms_caller = NULL;
static Counter *num_outstanding_requests;

static Octstr *interface;

/* how to contact the MMS server */
static Octstr *mmsscUrl = NULL;

static Octstr *wapIpAddress = NULL;
static long wapPort = 8201;

static Octstr *mmsDirectory;
static Octstr *mmsSignalFile = NULL;
static Octstr *mmsSignalURL = NULL;

static Counter *mms_client_counter;
static Counter *mms_transaction_counter;

static long mms_control_port;
static Octstr *mms_control_interface;
static Octstr *mms_control_url;

static long default_bearerinit_delay;

static unsigned long max_http_retries = HTTP_MAX_RETRIES;
static unsigned long http_queue_delay = HTTP_RETRY_DELAY;



/*
 * Private functions.
 */

static void main_thread(void *);
static void bearer_polling_thread(void *);
static void mms_control_thread(void *arg);
static void mms_retrieve_unanswered();
static void url_result_thread(void *arg);
static void http_queue_thread(void *arg);

/***********************************************************************
 * The public interface to the application layer.
 */

void mms_appl_init(Cfg *cfg) {
    CfgGroup *grp;

    gw_assert(run_status == limbo);

    queue = list_create();
    list_add_producer(queue);

    incoming_mms = list_create();
    list_add_producer(incoming_mms);

    clientMachines = list_create();
    list_add_producer(clientMachines);

    clientSendMachines = list_create();
    list_add_producer(clientSendMachines);

    http_requests = list_create();
    list_add_producer(http_requests);

    mms_client_counter = counter_create();
    mms_transaction_counter = counter_create();

    num_outstanding_requests = counter_create();

    run_status = running;
    mms_caller = http_caller_create();

    grp = cfg_get_single_group(cfg, octstr_imm("mms-client"));
    if (grp == NULL)
        panic(0, "No 'mms-client' group in configuration");


    /* notifying client of new MMS messages */
    mmsSignalURL = cfg_get(grp, octstr_imm("mms-signal-url"));
    mmsDirectory = cfg_get(grp, octstr_imm("mms-directory"));
    mmsSignalFile = cfg_get(grp, octstr_imm("mms-signal-file"));

    /* bearer interface (probably ppp0) for wap */
    interface = cfg_get(grp, octstr_imm("wap-bearer-interface"));

    /* information about the MMS service centre */
    mmsscUrl = cfg_get(grp, octstr_imm("mms-sc-url"));

    /* Information about the WAP relay host */
    cfg_get_integer(&wapPort, grp, octstr_imm("wap-port"));
    wapIpAddress = cfg_get(grp, octstr_imm("wap-ipaddress"));

    /* controlling the MMS stack */
    cfg_get_integer(&mms_control_port, grp, octstr_imm("mms-control-port"));
    if ((mms_control_url = cfg_get(grp, octstr_imm("mms-control-url"))) == NULL)
        mms_control_url = octstr_imm("/cgi-bin/mms");

    /* controlling the MMS stack */
    cfg_get_integer(&default_bearerinit_delay, grp, octstr_imm("bearer-init-timeout"));
    if (default_bearerinit_delay == -1)
        default_bearerinit_delay = 60;


    /* read any cached PDUs from disk and stick back in queue */
    mms_retrieve_unanswered();

    if (mms_control_port > 0) {
        if (http_open_port_if(mms_control_port, 0, mms_control_interface) == -1) {
            panic(0, "Failed to open HTTP socket");
        } else {
            info(0, "Set up MMS control service at port %ld", mms_control_port);
            gwthread_create(mms_control_thread, NULL);
            gwthread_join_every(mms_control_thread);
        }
    }

    gwthread_create(main_thread, cfg);
    gwthread_create(bearer_polling_thread, NULL);

    /* only need HTTP support if configured */
    if (mmsSignalURL != NULL || wapIpAddress == NULL) {
        gwthread_create(http_queue_thread, NULL);
        gwthread_create(url_result_thread, NULL);
    }
}


void mms_appl_shutdown(void) {
    gw_assert(run_status == running);
    run_status = terminating;

    counter_destroy(mms_client_counter);
    counter_destroy(mms_transaction_counter);

    counter_destroy(num_outstanding_requests);

    list_remove_producer(queue);
    list_remove_producer(clientMachines);
    list_remove_producer(clientSendMachines);
    list_remove_producer(incoming_mms);
    list_remove_producer(http_requests);

    gwthread_join_every(main_thread);
    gwthread_join_every(bearer_polling_thread);

    http_caller_signal_shutdown(mms_caller);
    /*gwthread_join_every(return_replies_thread);*/

    http_caller_destroy(mms_caller);

    list_destroy(queue, wap_event_destroy_item);

    gw_assert(list_len(incoming_mms) == 0);
    list_destroy(incoming_mms, NULL);

    list_destroy(http_requests, NULL);

    octstr_destroy(interface);
    octstr_destroy(mmsscUrl);
    octstr_destroy(mmsDirectory);
    octstr_destroy(mmsSignalFile);
    octstr_destroy(mmsSignalURL);
    octstr_destroy(mms_control_interface);
    octstr_destroy(mms_control_url);

}


void mms_appl_dispatch(WAPEvent *event) {
    gw_assert(run_status == running);
    list_produce(queue, event);
}



MMSClientSendMachine *mms_create_client_send_machine(WAPEvent *event) {
    MMSClientSendMachine *machine;
    machine = gw_malloc(sizeof(*machine));
    machine->mms_send_req = event;
    machine->sendconf = NULL;
    machine->state = state_idle;
    machine->timer = gwtimer_create(queue);
    machine->machine_id = counter_increase(mms_client_counter);
    return machine;
}

MMSClientMachine *mms_create_client_machine(MMS_PDU *pdu) {
    MMSClientMachine *machine;
    machine = gw_malloc(sizeof(*machine));
    machine->notification = pdu;
    machine->retrieveconf = NULL;
    machine->addrtuple = NULL;
    machine->state = state_idle;
    machine->timer = gwtimer_create(queue);
    machine->machine_id = counter_increase(mms_client_counter);

    /* TODO write PDU to disk */
    machine->savedNotification = NULL;

    return machine;
}

void mms_destroy_client_send_machine(MMSClientSendMachine *machine) {
    if (machine->mms_send_req != NULL)
        wap_event_destroy(machine->mms_send_req);
    if (machine->sendconf != NULL)
        mms_pdu_destroy(machine->sendconf);
    if (machine->timer != NULL)
        gwtimer_destroy(machine->timer);

    gw_free(machine);
}

void mms_destroy_client_machine(MMSClientMachine *machine) {
    if (machine->notification != NULL)
        mms_pdu_destroy(machine->notification);
    if (machine->retrieveconf != NULL)
        mms_pdu_destroy(machine->retrieveconf);
    if (machine->addrtuple != NULL)
        wap_addr_tuple_destroy(machine->addrtuple);
    if (machine->timer != NULL)
        gwtimer_destroy(machine->timer);

    gw_free(machine);
}


static char *state_name(int state) {
    switch (state) {
#define STATE_NAME(name) case name: return #name;
#define ROW(state, event, cond, stmt, next_state)
#include "mms/mms_receiver_states.def"

#define STATE_NAME(name) case name: return #name;
#define ROW(state, event, cond, stmt, next_state)
#include "mms/mms_sender_states.def"

        default:
            return "unknown mms state";
    }
}



static void mms_retrieve_unanswered() {
    /* TODO read from disk and put in queue */
}



struct receiver {
    unsigned long handle;
    int method;  /* the HTTP method to use */
    Octstr *url; /* the after pattern URL */
    List *http_headers; 
    Octstr *body; /* body content of the request */
    unsigned long retries; /* number of performed retries */
};


void *remember_receiver(unsigned long handle, int method, Octstr *url, List *headers, 
                               Octstr *body, unsigned int retries) {
    struct receiver *receiver;

    counter_increase(num_outstanding_requests);
    receiver = gw_malloc(sizeof(*receiver));

    receiver->handle = handle;

    receiver->method = method;
    receiver->url = octstr_duplicate(url);
    receiver->http_headers = http_header_duplicate(headers);
    receiver->body = octstr_duplicate(body);
    receiver->retries = retries;

    /* TODO save receiver to file */

    return receiver;
}



static void get_receiver(void *id, unsigned long *handle, int *method, Octstr **url, List **headers, Octstr **body, unsigned long *retries) {
    struct receiver *receiver;

    receiver = id;
    *handle = receiver->handle;
    *method = receiver->method;
    *url = receiver->url;
    *headers = receiver->http_headers;
    *body = receiver->body;
    *retries = receiver->retries;
    gw_free(receiver);
    counter_decrease(num_outstanding_requests);
}



static void notifyClientByHttp(Octstr *filename, long handle, Octstr *content) {
    List *headers;
    void *id;

    headers = http_create_empty_headers();
    http_header_add(headers, "User-Agent", GW_NAME "/" VERSION);
    http_header_add(headers, "X-MMS-Filename", octstr_get_cstr(filename));
    /* response will come back asynchronously */

    id = remember_receiver(handle, HTTP_METHOD_POST, mmsSignalURL, headers, content, 30);

    http_start_request(mms_caller, HTTP_METHOD_POST, mmsSignalURL, headers, content, 1, id, NULL);
}

static void *notifyClient(long status, MMS_PDU *pdu, MMSClientMachine *machine) {
    FILE *fp;
    Octstr *filename = NULL;
    Octstr *xml;

    /* some fields might need to be copied from the notification */
    if (pdu->u.RetrieveConf.from == NULL)
        pdu->u.RetrieveConf.from = octstr_duplicate(machine->notification->u.NotificationInd.from);
    if (pdu->u.RetrieveConf.subject == NULL)
        pdu->u.RetrieveConf.subject = octstr_duplicate(machine->notification->u.NotificationInd.subject);

    /* Save to file in XML format */
    if (status == HTTP_OK) {
        filename = octstr_format("%s%ld_%ld.xml", octstr_get_cstr(mmsDirectory), pdu->u.RetrieveConf.date, machine->machine_id);
        fp = fopen(octstr_get_cstr(filename), "w");
        xml = mms_pdu_to_xml(pdu);
        octstr_print(fp, xml);
        fclose(fp);
        octstr_destroy(xml);

        if (mmsSignalFile != NULL) {
            /* append filename to signal file */
            fp = fopen(octstr_get_cstr(mmsSignalFile), "a");
            octstr_print(fp, octstr_format("%s\n", octstr_get_cstr(filename)));
            fclose(fp);            
        }
        if (mmsSignalURL != NULL) {
            notifyClientByHttp(filename, machine->machine_id, xml);
            machine->filename = filename;
        }
    }
}




/* check whether the network interface is there */
static int checkInterface(Octstr *interface) {
    FILE *fh;
    char buf[512];
    int found = 0;
    Octstr *line;

    fh = fopen("/proc/net/dev", "r");
    if (fh) {
        // skipping first 2 lines
        fgets(buf, sizeof buf, fh); /* eat line */
        fgets(buf, sizeof buf, fh);

        while (! found && fgets(buf, sizeof buf, fh)) {
            line = octstr_create(buf);
            found = octstr_search(line, interface, 0) >= 0;
            octstr_destroy(line);
        }
        fclose(fh);
    }
    return found;
}


static int get_addr(char *ifname, struct sockaddr *ifaddr) {

    struct ifreq *ifr;
    struct ifreq ifrr;
    struct sockaddr_in sa;
    int sockfd;

    ifr = &ifrr;

    ifrr.ifr_addr.sa_family = AF_INET;

    strncpy(ifrr.ifr_name, ifname, sizeof(ifrr.ifr_name));

    if (0 > (sockfd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP))) {
        fprintf(stderr, "Cannot open socket.\n");
        return -2;
    }

    if (ioctl(sockfd, SIOCGIFADDR, ifr) < 0) {
        printf("No %s interface.\n", ifname);
        return -1;
    }

    *ifaddr = ifrr.ifr_addr;
    /*printf("Address for %s: %s\n", ifname, inet_ntoa(inaddrr(ifr_addr.sa_data)));*/

    close(sockfd);

    return 0;
}


WAPAddrTuple *getRequestTuple(Octstr *interface) {
    struct sockaddr localaddr;
    Octstr *localAddr;
    struct hostent hostinfo;

    /* get the correct local IP address */
    get_addr(octstr_get_cstr(interface), &localaddr);
    localAddr = gw_netaddr_to_octstr(AF_INET, (struct in_addr *) &localaddr.sa_data); 

    return wap_addr_tuple_create(wapIpAddress, wapPort, localAddr, 9201);
}



static void sendWSPConnectRequest(WAPAddrTuple *addrTuple, long handle) {
    WAPEvent *e;

    /* start connection */
    e = wap_event_create(S_Connect_Req);
    /* no client headers needed for MMS
       leave capability negotiation to default */
    e->u.S_Connect_Req.addr_tuple = wap_addr_tuple_duplicate(addrTuple);
    e->u.S_Connect_Req.session_handle = handle;
    wsp_client_session_dispatch_event(e);
}


static void sendWSPDisconnectRequest(long handle) {
    WAPEvent *e;

    /* start connection */
    e = wap_event_create(S_Disconnect_Req);
    e->u.S_Disconnect_Req.session_handle = handle;
    wsp_client_session_dispatch_event(e);
}

static void sendWSPMethodAbort(long transaction_id, long handle) {
    WAPEvent *e;

    /* start connection */
    e = wap_event_create(S_MethodAbort_Req);
    e->u.S_MethodAbort_Req.session_handle = handle;
    e->u.S_MethodAbort_Req.transaction_id = transaction_id;
    wsp_client_session_dispatch_event(e);
}


static void sendWspAcknowledgement(Octstr *transaction_id, MMSClientMachine *machine) {
    MMS_PDU *pdu;
    WAPEvent *e;

    pdu = mms_pdu_create(AcknowledgeInd);
    pdu->u.AcknowledgeInd.x_mms_transaction_id = octstr_duplicate(transaction_id);
    pdu->u.AcknowledgeInd.x_mms_mms_version = octstr_imm("1.0");

    machine->transaction_id = counter_increase(mms_transaction_counter);

    e = wap_event_create(S_MethodInvoke_Req);
    /* no client headers needed for MMS
       leave capability negotiation to default */
    e->u.S_MethodInvoke_Req.client_transaction_id = machine->transaction_id;
    e->u.S_MethodInvoke_Req.session_handle = machine->machine_id;
    e->u.S_MethodInvoke_Req.method = octstr_imm("POST");
    e->u.S_MethodInvoke_Req.request_uri = mmsscUrl;
    e->u.S_MethodInvoke_Req.request_body = mms_pdu_pack(pdu);
    wsp_client_session_dispatch_event(e);
}

static void sendWspNotifyResponse(Octstr *transaction_id, long status, MMSClientMachine *machine) {
    MMS_PDU *pdu;
    WAPEvent *e;

    pdu = mms_pdu_create(NotifyRespInd);
    pdu->u.NotifyRespInd.x_mms_transaction_id = octstr_duplicate(transaction_id);
    pdu->u.NotifyRespInd.x_mms_mms_version = octstr_imm("1.0");
    pdu->u.NotifyRespInd.x_mms_status = status;

    machine->transaction_id = counter_increase(mms_transaction_counter);

    e = wap_event_create(S_MethodInvoke_Req);
    /* no client headers needed for MMS
       leave capability negotiation to default */
    e->u.S_MethodInvoke_Req.client_transaction_id = machine->transaction_id;
    e->u.S_MethodInvoke_Req.session_handle = machine->machine_id;
    e->u.S_MethodInvoke_Req.method = octstr_imm("POST");
    e->u.S_MethodInvoke_Req.request_uri = mmsscUrl;
    e->u.S_MethodInvoke_Req.request_body = mms_pdu_pack(pdu);
    wsp_client_session_dispatch_event(e);
}


static void notifySendClientDeliveryReport() {
    /* TODO send delivery report to client */
}


MMS_PDU *createSendReqPDU(MMSClientSendMachine *machine) {
    MMS_PDU *pdu;
    time_t now;
    
    now = time(NULL);

    pdu = mms_pdu_create(SendReq);
    pdu->u.SendReq.x_mms_transaction_id =       octstr_duplicate(machine->mms_send_req->u.MMS_Message_Request.handle);
    pdu->u.SendReq.x_mms_mms_version =          octstr_imm("1.0");
    pdu->u.SendReq.date =                       (long) now;
    pdu->u.SendReq.from =                       octstr_duplicate(machine->mms_send_req->u.MMS_Message_Request.from);
    pdu->u.SendReq.from_charset =               machine->mms_send_req->u.MMS_Message_Request.from_charset;
    pdu->u.SendReq.to =                         octstr_duplicate(machine->mms_send_req->u.MMS_Message_Request.to);
    pdu->u.SendReq.to_charset =                 machine->mms_send_req->u.MMS_Message_Request.to_charset;
    pdu->u.SendReq.cc =                         octstr_duplicate(machine->mms_send_req->u.MMS_Message_Request.cc);
    pdu->u.SendReq.cc_charset =                 machine->mms_send_req->u.MMS_Message_Request.cc_charset;
    pdu->u.SendReq.bcc =                        octstr_duplicate(machine->mms_send_req->u.MMS_Message_Request.bcc);
    pdu->u.SendReq.bcc_charset =                machine->mms_send_req->u.MMS_Message_Request.bcc_charset;
    pdu->u.SendReq.subject =                    octstr_duplicate(machine->mms_send_req->u.MMS_Message_Request.subject);
    pdu->u.SendReq.x_mms_message_class =        machine->mms_send_req->u.MMS_Message_Request.message_class;
    pdu->u.SendReq.x_mms_expiry =               machine->mms_send_req->u.MMS_Message_Request.expiry;
    pdu->u.SendReq.x_mms_expiry_absrelind =     machine->mms_send_req->u.MMS_Message_Request.expiry_type;
    pdu->u.SendReq.x_mms_delivery_time =        machine->mms_send_req->u.MMS_Message_Request.delivery_time;
    pdu->u.SendReq.x_mms_delivery_time_absrelind = machine->mms_send_req->u.MMS_Message_Request.delivery_time_type;
    pdu->u.SendReq.x_mms_priority =             machine->mms_send_req->u.MMS_Message_Request.priority;
    pdu->u.SendReq.x_mms_sender_visibility =    machine->mms_send_req->u.MMS_Message_Request.sender_visibility;
    pdu->u.SendReq.x_mms_delivery_report =      machine->mms_send_req->u.MMS_Message_Request.delivery_report;
    pdu->u.SendReq.x_mms_read_reply =           machine->mms_send_req->u.MMS_Message_Request.read_reply;
    pdu->u.SendReq.content_type =               octstr_duplicate(machine->mms_send_req->u.MMS_Message_Request.content_type);
    pdu->u.SendReq.data =                       octstr_duplicate(machine->mms_send_req->u.MMS_Message_Request.message);

}

static void sendWSPMSendReq(MMSClientSendMachine *machine) {
    MMS_PDU *pdu;
    WAPEvent *e;

    machine->transaction_id = counter_increase(mms_transaction_counter);
    pdu = createSendReqPDU(machine);

    e = wap_event_create(S_MethodInvoke_Req);
    /* no client headers needed for MMS
       leave capability negotiation to default */
    e->u.S_MethodInvoke_Req.client_transaction_id = machine->transaction_id;
    e->u.S_MethodInvoke_Req.session_handle = machine->machine_id;
    e->u.S_MethodInvoke_Req.method = octstr_imm("POST");
    e->u.S_MethodInvoke_Req.request_uri = mmsscUrl;
    e->u.S_MethodInvoke_Req.request_body = mms_pdu_pack(pdu);
    wsp_client_session_dispatch_event(e);

    mms_pdu_destroy(pdu);
}






static void wspGet(Octstr *url, MMSClientMachine *machine) {
    WAPEvent *e;

    machine->transaction_id = counter_increase(mms_transaction_counter);

    e = wap_event_create(S_MethodInvoke_Req);
    /* no client headers needed for MMS
       leave capability negotiation to default */
    e->u.S_MethodInvoke_Req.client_transaction_id = machine->transaction_id;
    e->u.S_MethodInvoke_Req.session_handle = machine->machine_id;
    e->u.S_MethodInvoke_Req.method = octstr_imm("GET");
    e->u.S_MethodInvoke_Req.request_uri = octstr_duplicate(url);
    wsp_client_session_dispatch_event(e);

}

static void start_timerTO_A_bearer(Timer *timer, long session_handle, long delay) {
    WAPEvent *e;
    /* start connection */
    e = wap_event_create(TimerTO_A_bearer);
    /* no client headers needed for MMS
       leave capability negotiation to default */
    e->u.TimerTO_A_bearer.session_handle = session_handle;
    e->u.TimerTO_A_bearer.delay = delay;   
    gwtimer_start(timer, delay, e);
}


static void start_timerTO_Http_Client(Timer *timer, long session_handle, long delay) {
    WAPEvent *e;
    /* start connection */
    e = wap_event_create(TimerTO_Http_Client);
    /* no client headers needed for MMS
       leave capability negotiation to default */
    e->u.TimerTO_Http_Client.session_handle = session_handle;
    gwtimer_start(timer, delay, e);
}



/* MMS over HTTP */

static void sendHTTPMSendReq(MMSClientSendMachine *machine) {
    MMS_PDU *pdu;
    WAPEvent *e;
    List *headers;
    void *id;
    Octstr *body;

    machine->transaction_id = counter_increase(mms_transaction_counter);
    pdu = createSendReqPDU(machine);
    body = mms_pdu_pack(pdu);

    headers = http_create_empty_headers();
    http_header_add(headers, "User-Agent", GW_NAME "/" VERSION);
    http_header_add(headers, "Content-Type", "application/vnd.wap.mms-message");
    id = remember_receiver(machine->machine_id, HTTP_METHOD_POST, mmsscUrl, headers, body, 30);
    http_start_request(mms_caller, HTTP_METHOD_POST, mmsscUrl, headers, body, 1, id, NULL);

    mms_pdu_destroy(pdu);
}




static void httpGet(Octstr *url, MMSClientMachine *machine) {
    List *headers;
    void *id;

    headers = http_create_empty_headers();
    http_header_add(headers, "User-Agent", GW_NAME "/" VERSION);
    id = remember_receiver(machine->machine_id, HTTP_METHOD_GET, url, headers, NULL, 30);
    http_start_request(mms_caller, HTTP_METHOD_GET, url, headers, NULL, 1, id, NULL);
}


static void sendHttpAcknowledgement(Octstr *transaction_id, MMSClientMachine *machine) {

    MMS_PDU *pdu;
    Octstr *body;
    List *headers;
    void *id;

    pdu = mms_pdu_create(AcknowledgeInd);
    pdu->u.AcknowledgeInd.x_mms_transaction_id = octstr_duplicate(transaction_id);
    pdu->u.AcknowledgeInd.x_mms_mms_version = octstr_imm("1.0");
    body = mms_pdu_pack(pdu);

    headers = http_create_empty_headers();
    http_header_add(headers, "User-Agent", GW_NAME "/" VERSION);
    http_header_add(headers, "Content-Type", "application/vnd.wap.mms-message");

    id = remember_receiver(machine->machine_id, HTTP_METHOD_POST, mmsscUrl, headers, body, 30);
    http_start_request(mms_caller, HTTP_METHOD_POST, mmsscUrl, headers, body, 1, id, NULL);
}



static void sendHttpNotifyResponse(Octstr *transaction_id, long status, MMSClientMachine *machine) {

    MMS_PDU *pdu;
    Octstr *body;
    List *headers;
    void *id;

    pdu = mms_pdu_create(NotifyRespInd);
    pdu->u.NotifyRespInd.x_mms_transaction_id = octstr_duplicate(transaction_id);
    pdu->u.NotifyRespInd.x_mms_mms_version = octstr_imm("1.0");
    pdu->u.NotifyRespInd.x_mms_status = status;
    body = mms_pdu_pack(pdu);

    headers = http_create_empty_headers();
    http_header_add(headers, "User-Agent", GW_NAME "/" VERSION);
    http_header_add(headers, "Content-Type", "application/vnd.wap.mms-message");

    id = remember_receiver(machine->machine_id, HTTP_METHOD_POST, mmsscUrl, headers, body, 30);
    http_start_request(mms_caller, HTTP_METHOD_POST, mmsscUrl, headers, body, 1, id, NULL);
}






static int find_client_by_machine_id(void *mms_ptr, void *id_ptr) {
    MMSClientMachine *mms = mms_ptr;
    long *idp = id_ptr;

    return mms->machine_id == *idp;
}

static int find_client_send_by_machine_id(void *mms_ptr, void *id_ptr) {
    MMSClientSendMachine *mms = mms_ptr;
    long *idp = id_ptr;

    return mms->machine_id == *idp;
}

static int find_client_send_by_message_id(void *mms_ptr, void *id_ptr) {
    MMSClientSendMachine *mms = mms_ptr;
    Octstr *idp = id_ptr;

    return octstr_compare(mms->sendconf->u.SendConf.message_id, idp) == 0;
}



static MMSClientMachine *findClientMachine(WAPEvent *event, MMS_PDU *pdu) {
    long machine_id;
    MMSClientMachine *machine = NULL;

    switch (event->type) {
        case S_Unit_Push_Ind:
            if (pdu->type == NotificationInd) {
                /* machine that will handle this */
                machine = mms_create_client_machine(pdu);
                list_append(clientMachines, machine);
            }
            break;
        case S_Connect_Cnf:
            machine_id = event->u.S_Connect_Cnf.session_handle;
            break;
        case S_Disconnect_Ind:
            machine_id = event->u.S_Disconnect_Ind.session_handle;
            break;
        case S_MethodInvoke_Cnf:
            machine_id = event->u.S_MethodInvoke_Cnf.session_handle;
            break;
        case S_MethodResult_Ind:
            machine_id = event->u.S_MethodResult_Ind.session_handle;
            break;
        case S_MethodAbort_Ind:
            machine_id = event->u.S_MethodAbort_Ind.session_handle;
            break;
        case TimerTO_A_bearer:
            machine_id = event->u.TimerTO_A_bearer.session_handle;
            break;
        case Bearer_Event:
            machine_id = event->u.Bearer_Event.session_handle;
            break;
        case HTTP_Request_Completed_Event:
            machine_id = event->u.HTTP_Request_Completed_Event.session_handle;
            break;
    }

    if (machine == NULL)
        machine = list_search(clientMachines, &machine_id, find_client_by_machine_id);
    return machine;
}



static MMSClientSendMachine *findClientSendMachine(WAPEvent *event, MMS_PDU *pdu) {
    long machine_id = -1;
    Octstr *message_id;
    MMSClientSendMachine *machine = NULL;

    switch (event->type) {
        case S_Unit_Push_Ind:
            if (pdu->type == DeliveryInd)
                message_id = pdu->u.DeliveryInd.message_id;
            break;
        case S_Connect_Cnf:
            machine_id = event->u.S_Connect_Cnf.session_handle;
            break;
        case S_Disconnect_Ind:
            machine_id = event->u.S_Disconnect_Ind.session_handle;
            break;
        case S_MethodInvoke_Cnf:
            machine_id = event->u.S_MethodInvoke_Cnf.session_handle;
            break;
        case S_MethodResult_Ind:
            machine_id = event->u.S_MethodResult_Ind.session_handle;
            break;
        case S_MethodAbort_Ind:
            machine_id = event->u.S_MethodAbort_Ind.session_handle;
            break;
        case TimerTO_A_bearer:
            machine_id = event->u.TimerTO_A_bearer.session_handle;
            break;
        case Bearer_Event:
            machine_id = event->u.Bearer_Event.session_handle;
            break;
        case MMS_Message_Request:
            machine = mms_create_client_send_machine(event);
            list_append(clientSendMachines, machine);
            break;
        case HTTP_Request_Completed_Event:
            machine_id = event->u.HTTP_Request_Completed_Event.session_handle;
            break;
    }

    if (machine == NULL)
        machine = list_search(clientSendMachines, &machine_id, find_client_send_by_machine_id);

    if (machine == NULL)
        machine = list_search(clientSendMachines, &message_id, find_client_send_by_message_id);

    return machine;
}



void startbearerservice() {
    if (modemConn != NULL)
        startdialup(NULL, modemConn);
}


void stopbearerservice() {
    if (modemConn != NULL)
        stopdialup(NULL, modemConn);
}


/* send to every active machine */
static void sendBearerEvent() {
    MMSClientMachine *clientMachine;
    MMSClientSendMachine *clientSendMachine;
    WAPEvent *e;
    long i;

    for (i = 0; i < list_len(clientMachines); i++) {
        clientMachine = list_get(clientMachines, i);
        e = wap_event_create(Bearer_Event);
        e->u.Bearer_Event.session_handle = clientMachine->machine_id;
        list_insert(queue, 0, e);
    }

    for (i = 0; i < list_len(clientSendMachines); i++) {
        clientSendMachine = list_get(clientSendMachines, i);    
        e = wap_event_create(Bearer_Event);
        e->u.Bearer_Event.session_handle = clientSendMachine->machine_id;
        list_insert(queue, 0, e);
    }
}

/* send to every active machine */
static void sendBearerLostEvent() {
    MMSClientMachine *clientMachine;
    MMSClientSendMachine *clientSendMachine;
    WAPEvent *e;
    long i;

    for (i = 0; i < list_len(clientMachines); i++) {
        clientMachine = list_get(clientMachines, i);    
        e = wap_event_create(Bearer_Lost_Event);
        e->u.Bearer_Lost_Event.session_handle = clientMachine->machine_id;
        list_insert(queue, 0, e);
    }

    for (i = 0; i < list_len(clientSendMachines); i++) {
        clientSendMachine = list_get(clientSendMachines, i);    
        e = wap_event_create(Bearer_Lost_Event);
        e->u.Bearer_Lost_Event.session_handle = clientSendMachine->machine_id;
        list_insert(queue, 0, e);
    }
}


static void clientHttpResponse(HTTPClient *client, MMS_PDU *pdu) {
    Octstr *answer;
    List *reply_hdrs;
    int status;

    reply_hdrs = http_create_empty_headers();
    http_header_add(reply_hdrs, "Pragma", "no-cache");
    http_header_add(reply_hdrs, "Cache-Control", "no-cache");

    if (pdu == NULL) {
        /* no response */
        http_header_add(reply_hdrs, "Content-type", "text/html");
        answer = octstr_create("Timed Out");
        status = HTTP_NOT_FOUND;
    } else {
        http_header_add(reply_hdrs, "Content-type", "text/xml");
        answer = mms_pdu_to_xml(pdu);
        status = HTTP_OK;
    }

    http_send_reply(client, status, reply_hdrs, answer);
    octstr_destroy(answer);
    http_destroy_headers(reply_hdrs);
}


static void clientHttpInform(Octstr *responseUrl, MMS_PDU *pdu, MMSClientSendMachine *machine) {
    List *headers;
    Octstr *body;

    headers = http_create_empty_headers();
    http_header_add(headers, "User-Agent", GW_NAME "/" VERSION);    
    http_header_add(headers, "Content-type", "text/xml");
    body = mms_pdu_to_xml(pdu);

    /* response will come back asynchronously */
    http_start_request(mms_caller, HTTP_METHOD_GET, mmsSignalURL, headers, body, 1, &machine->machine_id, NULL);
}



static void notifySendClient(HTTPClient *client, Octstr *responseUrl, MMS_PDU *pdu, MMSClientSendMachine *machine) {
    if (client != NULL)
        clientHttpResponse(client, pdu);
    else
        clientHttpInform(responseUrl, pdu, machine);
}



static int handle_mms_request(HTTPClient *client, List *headers, Octstr *body, List *args) {

    WAPEvent *e;
    Octstr *tmp_string;

    e = wap_event_create(MMS_Message_Request);

    /* read arguments from web into event */
    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "handle");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%ld", &e->u.MMS_Message_Request.handle);

    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "timeout");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%ld", &e->u.MMS_Message_Request.timeout);
    else
        /* default is 15 minutes */
        e->u.MMS_Message_Request.timeout = 900;

    e->u.MMS_Message_Request.from = octstr_duplicate(http_cgi_variable(args, "from"));
    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "from-charset");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%d", &e->u.MMS_Message_Request.from_charset);

    e->u.MMS_Message_Request.to = octstr_duplicate(http_cgi_variable(args, "to"));
    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "to-charset");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%d", &e->u.MMS_Message_Request.to_charset);

    e->u.MMS_Message_Request.cc = octstr_duplicate(http_cgi_variable(args, "cc"));
    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "cc-charset");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%d", &e->u.MMS_Message_Request.cc_charset);

    e->u.MMS_Message_Request.bcc = octstr_duplicate(http_cgi_variable(args, "bcc"));
    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "bcc-charset");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%d", &e->u.MMS_Message_Request.bcc_charset);

    e->u.MMS_Message_Request.subject = octstr_duplicate(http_cgi_variable(args, "subject"));
    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "subject-charset");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%d", &e->u.MMS_Message_Request.subject_charset);

    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "message-class");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%d", &e->u.MMS_Message_Request.message_class);

    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "expiry");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%ld", &e->u.MMS_Message_Request.expiry);
    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "expiry-type");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%d", &e->u.MMS_Message_Request.expiry_type);

    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "delivery-time");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%ld", &e->u.MMS_Message_Request.delivery_time);
    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "delivery-time-type");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%d", &e->u.MMS_Message_Request.delivery_time_type);

    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "priority");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%d", &e->u.MMS_Message_Request.priority);

    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "sender-visibility");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%d", &e->u.MMS_Message_Request.sender_visibility);

    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "delivery-report");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%d", &e->u.MMS_Message_Request.delivery_report);

    tmp_string = NULL;
    tmp_string = http_cgi_variable(args, "read-reply");
    if (tmp_string != NULL)
        sscanf(octstr_get_cstr(tmp_string),"%d", &e->u.MMS_Message_Request.read_reply);

    e->u.MMS_Message_Request.content_type = octstr_duplicate(http_cgi_variable(args, "content-type"));

    e->u.MMS_Message_Request.message = octstr_duplicate(http_cgi_variable(args, "message"));

    e->u.MMS_Message_Request.delivery_notification_url = octstr_duplicate(http_cgi_variable(args, "delivery-notification-url"));

    e->u.MMS_Message_Request.response_url = octstr_duplicate(http_cgi_variable(args, "response-url"));

    if (e->u.MMS_Message_Request.response_url == NULL)
        /* block until the response is available */
        e->u.MMS_Message_Request.http_client = client;

    /* queue the request */
    list_produce(queue, e);

    return e->u.MMS_Message_Request.response_url != NULL;
}



static void handle_client_machine(MMSClientMachine *sm, WAPEvent *current_event, MMS_PDU *pdu) {

    debug("wap.mms", 0, "MMS: machine %p, state %s, event %s", (void *) sm, state_name(sm->state), wap_event_name(current_event->type));

#define STATE_NAME(name)
#define ROW(state_name, event, condition, action, next_state) \
		{ \
			struct event *e; \
			e = &current_event->u.event; \
			if (sm->state == state_name && \
			   current_event->type == event && \
			   (condition)) { \
				action \
				sm->state = next_state; \
				debug("mms.wsp", 0, "WSP %ld: New state %s", \
					sm->machine_id, #next_state); \
				goto end; \
			} \
		}
#include "mms/mms_receiver_states.def"

    end:
    wap_event_destroy(current_event);

    if (sm->state == NULL_STATE)
        mms_destroy_client_machine(sm);
}

static void handle_client_send_machine(MMSClientSendMachine *sm, WAPEvent *current_event, MMS_PDU *pdu) {

    debug("wap.mms", 0, "MMS: machine %p, state %s, event %s", (void *) sm, state_name(sm->state), wap_event_name(current_event->type));

#define STATE_NAME(name)
#define ROW(state_name, event, condition, action, next_state) \
		{ \
			struct event *e; \
			e = &current_event->u.event; \
			if (sm->state == state_name && \
			   current_event->type == event && \
			   (condition)) { \
				action \
				sm->state = next_state; \
				debug("mms.wsp", 0, "WSP %ld: New state %s", \
					sm->machine_id, #next_state); \
				goto end; \
			} \
		}
#include "mms/mms_sender_states.def"

    end:
    if (current_event->type != MMS_Message_Request)     /* we're hanging onto the message request */
        wap_event_destroy(current_event);

    if (sm->state == NULL_STATE)
        mms_destroy_client_send_machine(sm);
}




static void mms_control_thread(void *arg) {
    HTTPClient *client;
    Octstr *ip, *url, *body, *answer;
    List *hdrs, *args, *reply_hdrs;
    int status;

    for (;;) {
        client = http_accept_request(mms_control_port, &ip, &url, &hdrs, &body, &args);

        if (client == NULL)
            break;

        reply_hdrs = http_create_empty_headers();
        http_header_add(reply_hdrs, "Content-type", "text/html");
        http_header_add(reply_hdrs, "Pragma", "no-cache");
        http_header_add(reply_hdrs, "Cache-Control", "no-cache");

        info(0, "mms-appl: Got HTTP request <%s> from <%s>", octstr_get_cstr(url), octstr_get_cstr(ip));

        /*
         * determine which kind of HTTP request this is any
         * call the necessary routine for it
         */

        /* dialup control */
        if (octstr_compare(url, mms_control_url) == 0) {
            if (handle_mms_request(client, hdrs, body, args)) {
                /* asynchronous, return a response immediately, there will be a callback
                   when the result comes in */
                status = HTTP_OK;
                http_send_reply(client, status, reply_hdrs, answer);
            }
        } else {
            answer = octstr_create("Unknown request.");
            status = HTTP_NOT_FOUND;
            http_send_reply(client, status, reply_hdrs, answer);
            octstr_destroy(answer);
        }

        octstr_destroy(ip);
        octstr_destroy(url);
        http_destroy_headers(hdrs);
        octstr_destroy(body);
        http_destroy_cgiargs(args);
        http_destroy_headers(reply_hdrs);
    }
}




static void bearer_polling_thread(void *arg) {
    Octstr *event;
    int prev_bearer_status = 0, current_bearer_status;
    int polling, i, j;
    WAPEvent *e;

    /* keep it simple, check the bearer service regularly and
       inform any active state machines of changes */
    while (run_status == running) {
        sleep(BEARER_POLLING_INTERVAL);
        current_bearer_status = checkInterface(interface);
        if (prev_bearer_status == 0 && current_bearer_status == 1) {
            sendBearerEvent();
        } else if (prev_bearer_status == 1 && current_bearer_status == 0) {
            sendBearerLostEvent();
        }
        if (current_bearer_status == 1) {
            int bearerneeded = 0;
            /* check how many state machines still need a bearer */
            i = list_len(clientMachines);
            /*
            if (i > 0) {
                j = 0;
                while ((!bearerneeded) && j < i)
                    bearerneeded = ((MMSClientMachine *)list_get(clientMachines, j))->state != CLIENT_WAITING;
            } */
            if (i > 0)
                bearerneeded = 1;
            if (!bearerneeded) {
                i = list_len(clientSendMachines);
                /*
                if (i > 0) {
                    j = 0;
                    while ((!bearerneeded) && j < i)
                        bearerneeded = ((MMSClientSendMachine *)list_get(clientSendMachines, j))->state != SEND_DELIVERY_WAITING;
                } */
                if (i > 0)
                    bearerneeded = 1;
            }
            if (!bearerneeded)
                /* bearer no longer needed */
                stopbearerservice();
        }
        prev_bearer_status = current_bearer_status;
    }
}




/***********************************************************************
 * Thread to handle failed HTTP requests and retries to deliver the 
 * information to the HTTP server. The thread uses the http_requests
 * queue that is spooled by url_result_thread in case the HTTP requests
 * fails.
 */

static void http_queue_thread(void *arg) {
    void *id;
    Msg *msg;
    Octstr *req_url;
    List *req_headers;
    Octstr *req_body;
    unsigned long handle, retries;
    int method;

    while ((id = list_consume(http_requests)) != NULL) {
        /* 
         * Sleep for a while in order not to block other operting requests.
         * Defaults to 10 sec. if not given via http-queue-delay directive in
         * smsbox group.
         */
        if (http_queue_delay > 0)
            gwthread_sleep(http_queue_delay);

        debug("Mms.http",0,"HTTP: Queue contains %ld outstanding requests",  
              list_len(http_requests));

        /* 
         * Get all required HTTP request data from the queue and reconstruct 
         * the id pointer for later lookup in url_result_thread.
         */
        get_receiver(id, &handle, &method, &req_url, &req_headers, &req_body, &retries);

        if (retries < max_http_retries) {
            id = remember_receiver(handle, method, req_url, req_headers, req_body, ++retries);

            debug("sms.http",0,"HTTP: Retrying request <%s> (%ld/%ld)",  
                  octstr_get_cstr(req_url), retries, max_http_retries);

            /* re-queue this request to the HTTPCaller list */
            http_start_request(mms_caller, method, req_url, req_headers, req_body, 1, id, NULL);
        }

        msg_destroy(msg);
        octstr_destroy(req_url);
        http_destroy_headers(req_headers);
        octstr_destroy(req_body);
    }
}



static void url_result_thread(void *arg) {
    Octstr *final_url, *reply_body;
    List *reply_headers;
    int status, method;
    Octstr *req_url;
    List *req_headers;
    Octstr *req_body;
    void *id;
    unsigned long handle, retries;
    MMSClientSendMachine *sendmachine = NULL;
    MMSClientMachine *machine = NULL;
    unsigned int queued; /* indicate if processes reply is requeued */

    for (;;) {
        id = http_receive_result(mms_caller, &status, &final_url, &reply_headers, &reply_body);

        if (id == NULL)
            break;

        debug("mms", 0, "http response %d (%d retries) received for %s", status, retries, octstr_get_cstr(final_url));

        get_receiver(id, &handle, &method, &req_url, &req_headers, &req_body, &retries);

        if (http_status_class(status) == HTTP_STATUS_SUCCESSFUL) {
            /* feed back to state machine */
            WAPEvent *e;
            e = wap_event_create(HTTP_Request_Completed_Event);
            e->u.HTTP_Request_Completed_Event.session_handle = handle;
            e->u.HTTP_Request_Completed_Event.status = status;
            e->u.HTTP_Request_Completed_Event.response_body = reply_body;
            reply_body = NULL;
            list_produce(queue, e);
        } 
        else if (retries > 0) {
            id = remember_receiver(handle, method, req_url, req_headers, req_body, retries);
            list_produce(http_requests, id);
            queued++;
            goto requeued;
        } 
        else {
            /* exhausted all the retries, get back to state machine */
            WAPEvent *e;
            e = wap_event_create(HTTP_Request_Completed_Event);
            e->u.HTTP_Request_Completed_Event.session_handle = handle;
            e->u.HTTP_Request_Completed_Event.status = status;
            e->u.HTTP_Request_Completed_Event.response_body = octstr_create("");
            list_produce(queue, e);           
            error(0, "unable to complete HTTP request %s", final_url);
        }
    }

    requeued:
    octstr_destroy(final_url);
    http_destroy_headers(reply_headers);
    octstr_destroy(reply_body);
    octstr_destroy(req_url);
    http_destroy_headers(req_headers);
    octstr_destroy(req_body);
}




static void main_thread(void *arg) {
    WAPEvent *e, *res;
    MMS_PDU *pdu;
    long sid;
    WAPAddrTuple *tuple;
    Cfg *cfg;
    MMSClientMachine *clientMachine;
    MMSClientSendMachine *clientSendMachine;

    cfg = (Cfg *)arg;


    while (run_status == running && (e = list_consume(queue)) != NULL) {

        pdu = NULL;

        /* Pdu needs to be unpacked in certain cases */
        if (e->type ==  S_Unit_Push_Ind) {
            pdu = mms_pdu_unpack(e->u.S_Unit_Push_Ind.push_body);
            if (pdu == NULL) {
                warning(0, "MMS: Broken PDU ignored.");
                wap_event_destroy(e);
            }
        }
        if (e->type == S_MethodResult_Ind) {
            pdu = mms_pdu_unpack(e->u.S_MethodResult_Ind.response_body);
            if (pdu == NULL) {
                warning(0, "MMS: Broken PDU ignored.");
                wap_event_destroy(e);
            }
        }
        if (e->type == HTTP_Request_Completed_Event && e->u.HTTP_Request_Completed_Event.status != HTTP_NO_CONTENT) {

            octstr_dump(e->u.HTTP_Request_Completed_Event.response_body, 0);
            pdu = mms_pdu_unpack(e->u.HTTP_Request_Completed_Event.response_body);
            if (pdu == NULL) {
                warning(0, "MMS: Broken PDU ignored.");
                wap_event_destroy(e);
            }
        }

        clientMachine = findClientMachine(e, pdu);
        if (clientMachine != NULL) {
            handle_client_machine(clientMachine, e, pdu);
            if (pdu != clientMachine->notification && pdu != clientMachine->retrieveconf)
                /* no need to hang on */
                mms_pdu_destroy(pdu);
        } else {
            clientSendMachine = findClientSendMachine(e, pdu);
            if (clientSendMachine != NULL) {
                handle_client_send_machine(clientSendMachine, e, pdu);
                if (pdu != clientSendMachine->sendconf)
                    mms_pdu_destroy(pdu);
            } else
                wap_event_destroy(e);
        }
    }

}





